/*
 * (C) Copyright 2000
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <ppcboot.h>
#include "cma302.h"

flash_info_t	flash_info[CFG_MAX_FLASH_BANKS]; /* info for FLASH chips	*/

/*-----------------------------------------------------------------------
 * Functions
 */
static int write_buff (flash_info_t *info, uchar *src, ulong addr, ulong cnt);
static int write_word (flash_info_t *info, ulong dest, ulong data);
static int  flash_protect (int flag, ulong from, ulong to, flash_info_t *info);

/*-----------------------------------------------------------------------
 * Protection Flags:
 */
#define FLAG_PROTECT_SET	0x01
#define FLAG_PROTECT_CLEAR	0x02

/*-----------------------------------------------------------------------
 */


/*
 * probe for the existence of flash at address "addr"
 * 0 = yes, 1 = bad Manufacturer's Id, 2 = bad Device Id
 */
static int
flash_probe_word(flash_addr_t addr)
{
	/* reset the flash */
	*addr = BANK_CMD_RST;

	/* check the manufacturer id */
	*addr = BANK_CMD_RD_ID;
	if (*BANK_ADDR_MAN(addr) != BANK_RD_ID_MAN)
		return 1;

	/* check the device id */
	*addr = BANK_CMD_RD_ID;
	if (*BANK_ADDR_DEV(addr) != BANK_RD_ID_DEV)
		return 2;

#ifdef FLASH_DEBUG
	printf("\nMaster Lock Config = 0x%08lx\n", *BANK_ADDR_CFGM(addr));
	printf("Block  0 Lock Config = 0x%08lx\n", *BANK_ADDR_CFG0(addr));
	printf("Block  1 Lock Config = 0x%08lx\n", *BANK_ADDR_CFG1(addr));
	printf("Block  2 Lock Config = 0x%08lx\n", *BANK_ADDR_CFG2(addr));
	printf("Block  3 Lock Config = 0x%08lx\n", *BANK_ADDR_CFG3(addr));
	printf("Block  4 Lock Config = 0x%08lx\n", *BANK_ADDR_CFG4(addr));
	printf("Block  5 Lock Config = 0x%08lx\n", *BANK_ADDR_CFG5(addr));
	printf("Block  6 Lock Config = 0x%08lx\n", *BANK_ADDR_CFG6(addr));
	printf("Block  7 Lock Config = 0x%08lx\n", *BANK_ADDR_CFG7(addr));
	printf("Block  8 Lock Config = 0x%08lx\n", *BANK_ADDR_CFG8(addr));
	printf("Block  9 Lock Config = 0x%08lx\n", *BANK_ADDR_CFG9(addr));
	printf("Block 10 Lock Config = 0x%08lx\n", *BANK_ADDR_CFG10(addr));
	printf("Block 11 Lock Config = 0x%08lx\n", *BANK_ADDR_CFG11(addr));
	printf("Block 12 Lock Config = 0x%08lx\n", *BANK_ADDR_CFG12(addr));
	printf("Block 13 Lock Config = 0x%08lx\n", *BANK_ADDR_CFG13(addr));
	printf("Block 14 Lock Config = 0x%08lx\n", *BANK_ADDR_CFG14(addr));
	printf("Block 15 Lock Config = 0x%08lx\n", *BANK_ADDR_CFG15(addr));
#endif

	/* reset the flash again */
	*addr = BANK_CMD_RST;

	return 0;
}

unsigned long
flash_init(void)
{
	flash_addr_t addr, eaddr;
	unsigned long total;
	int i, nbanks;

	/* Init: no FLASHes known */
	for (i=0; i<CFG_MAX_FLASH_BANKS; ++i) {
		flash_info[i].flash_id = FLASH_UNKNOWN;
	}

	total = 0L;
	nbanks = 0;

	addr = (flash_addr_t)CFG_FLASH_BASE;
	eaddr = BANK_BASE_ADDR(addr, MAX_BANKS);

	while (addr < eaddr) {
		flash_addr_t addrw, eaddrw, addrb;

		addrw = addr;
		eaddrw = BANK_NEXT_WORD_ADDR(addrw);

		while (addrw < eaddrw)
			if (flash_probe_word(addrw++) != 0)
				goto out;

		if (nbanks >= CFG_MAX_FLASH_BANKS) {
			printf("[more flash available than config allows!] ");
			break;
		}

		/* fill in info for this bank */
		flash_info[nbanks].size = BANK_SIZE;
		flash_info[nbanks].sector_count = BANK_NBLOCKS;
		flash_info[nbanks].flash_id = FLASH_MAN_INTEL|FLASH_28F008S5;
		addrb = addr;
		for (i = 0; i < BANK_NBLOCKS; i++) {
			flash_info[nbanks].start[i] = (ulong)addrb;
			flash_info[nbanks].protect[i] = 0;
			addrb = BANK_NEXT_BLOCK_ADDR(addrb);
		}

		total += BANK_SIZE;
		nbanks++;

		addr = BANK_NEXT_BANK_ADDR(addr);
	}

out:
	if (nbanks == 0) {
		printf(
	"\nERROR: flash_init: no flash found at address 0x%08lx\n",
			(unsigned long)CFG_FLASH_BASE);
		hang();
	}

	/* protect monitor and environment */
#if CFG_MONITOR_BASE == CFG_FLASH_BASE
	(void)flash_protect(FLAG_PROTECT_SET,
		      CFG_MONITOR_BASE,
		      CFG_MONITOR_BASE+CFG_MONITOR_LEN-1,
		      &flash_info[0]);
#endif

#if defined(CFG_FLASH_ENV_ADDR)
	(void)flash_protect(FLAG_PROTECT_SET,
		      CFG_FLASH_ENV_ADDR,
#if defined(CFG_FLASH_ENV_BUF)
		      CFG_FLASH_ENV_ADDR + CFG_FLASH_ENV_BUF - 1,
#else
		      CFG_FLASH_ENV_ADDR + CFG_FLASH_ENV_SIZE - 1,
#endif
		      &flash_info[0]);
#endif

	return total;
}

/*-----------------------------------------------------------------------
 * Check or set protection status for monitor sectors
 *
 * The monitor always occupies the _first_ part of the _first_ Flash bank.
 */
static int
flash_protect(int flag, ulong from, ulong to, flash_info_t *info)
{
	ulong b_end = info->start[0] + info->size - 1;	/* bank end address */
	int rc    =  0;
	int first = -1;
	int last  = -1;
	int i;

	if (to < info->start[0]) {
		return (0);
	}

	for (i=0; i<info->sector_count; ++i) {
		ulong end;		/* last address in current sect	*/
		short s_end;

		s_end = info->sector_count - 1;

		end = (i == s_end) ? b_end : info->start[i + 1] - 1;

		if (from > end) {
			continue;
		}
		if (to < info->start[i]) {
			continue;
		}

		if (from == info->start[i]) {
			first = i;
			if (last < 0) {
				last = s_end;
			}
		}
		if (to  == end) {
			last  = i;
			if (first < 0) {
				first = 0;
			}
		}
	}

	for (i=first; i<=last; ++i) {
		if (flag & FLAG_PROTECT_CLEAR) {
			info->protect[i] = 0;
		} else if (flag & FLAG_PROTECT_SET) {
			info->protect[i] = 1;
		}
		if (info->protect[i]) {
			rc = 1;
		}
	}
	return (rc);
}


/*-----------------------------------------------------------------------
 */
void
flash_print_info(flash_info_t *info)
{
	int i;

	if (info->flash_id == FLASH_UNKNOWN) {
		printf ("missing or unknown FLASH type\n");
		return;
	}

	switch (info->flash_id & FLASH_VENDMASK) {
	case FLASH_MAN_INTEL:	printf ("INTEL ");		break;
	default:		printf ("Unknown Vendor ");	break;
	}

	switch (info->flash_id & FLASH_TYPEMASK) {
	case FLASH_28F008S5:	printf ("28F008S5\n");
				break;
	default:		printf ("Unknown Chip Type\n");
				break;
	}

	printf ("  Size: %ld MB in %d Sectors\n",
		info->size >> 20, info->sector_count);

	printf ("  Sector Start Addresses:");
	for (i=0; i<info->sector_count; ++i) {
		if ((i % 4) == 0)
			printf ("\n   ");
		printf (" %2d - %08lX%s", i,
			info->start[i],
			info->protect[i] ? " (RO)" : "     "
		);
	}
	printf ("\n");
}

/*-----------------------------------------------------------------------
 */


/*-----------------------------------------------------------------------
 */

/*
 * The following code cannot be run from FLASH!
 */

void
flash_erase(flash_info_t *info, int s_first, int s_last, bd_t *bd)
{
	int flag, prot, sect, errcount;
	ulong start, now, last;
#if defined(CFG_CMA302_PARALLEL_ERASE)
	int alldone;
	uchar done[CFG_MAX_FLASH_SECT];
#endif

#ifdef FLASH_DEBUG
	printf("\nflash_erase: erase %d sectors (%d to %d incl.) from\n"
		"  Bank # %d: ", s_last - s_first + 1, s_first, s_last,
		(info - flash_info) + 1);
	flash_print_info(info);
	printf("0x%08lx BANK_CMD_PROG\n", BANK_CMD_PROG);
	printf("0x%08lx BANK_CMD_ERASE1\n", BANK_CMD_ERASE1);
	printf("0x%08lx BANK_CMD_ERASE2\n", BANK_CMD_ERASE2);
	printf("0x%08lx BANK_CMD_CLR_STAT\n", BANK_CMD_CLR_STAT);
	printf("0x%08lx BANK_CMD_RST\n", BANK_CMD_RST);
	printf("0x%08lx BANK_STAT_RDY\n", BANK_STAT_RDY);
	printf("0x%08lx BANK_STAT_ERR\n", BANK_STAT_ERR);
#endif

	if ((s_first < 0) || (s_first > s_last)) {
		if (info->flash_id == FLASH_UNKNOWN) {
			printf ("- missing\n");
		} else {
			printf ("- no sectors to erase\n");
		}
		return;
	}

	if (info->flash_id != (FLASH_MAN_INTEL|FLASH_28F008S5)) {
		printf ("Flash type must be Intel 28F008S5 - aborted\n");
		return;
	}

	prot = 0;
	for (sect=s_first; sect<=s_last; ++sect) {
		if (info->protect[sect]) {
			prot++;
		}
	}

	if (prot) {
		printf("- Warning: %d protected sector%s will not be erased!\n",
			prot, (prot > 1 ? "s" : ""));
	}

#if defined(CFG_CMA302_PARALLEL_ERASE)

	/*
	 * erase all the unprotected sectors in parallel
	 *
	 * Quote from the Intel 28F008S5 manual, section 4.5 "Block
	 * Erase Command", "Erase is executed one block at a time ...".
	 *
	 * I take this to mean that you cannot erase multiple blocks
	 * in parallel. But I'll leave this code here for reference...
	 */

	/* Disable interrupts which might cause a timeout here */
	flag = disable_interrupts();

	/* Start erase on unprotected sectors */
	for (sect = s_first; sect<=s_last; sect++) {
		if (info->protect[sect] == 0) {	/* not protected */
			flash_addr_t addrw, eaddrw;

			addrw = (flash_addr_t)info->start[sect];
			eaddrw = BANK_NEXT_WORD_ADDR(addrw);

			while (addrw < eaddrw) {
				*addrw = BANK_CMD_ERASE1;
				*addrw = BANK_CMD_ERASE2;
				addrw++;
			}

			done[sect] = 0;
		}
	}

	/* re-enable interrupts if necessary */
	if (flag)
		enable_interrupts();

	/* wait at least 80us - let's wait 1 ms */
	udelay (1000, bd->bi_intfreq);

	start = get_timer (0);
	last = 0;
	errcount = 0;

	do {
		if ((now = get_timer(start)) > CFG_FLASH_ERASE_TOUT) {
			printf ("Timeout\n");
			return;
		}
		/* show that we're waiting */
		if ((now - last) > 1000) {	/* every second */
			serial_putc ('.');
			last = now;
		}

		/* see if all sectors are finished */
		alldone = 1;

		for (sect = s_first; sect <= s_last; sect++) {
			if (info->protect[sect] == 0 && !done[sect]) {
				flash_addr_t addrw, eaddrw;
				int sectdone, secterr;

				addrw = (flash_addr_t)info->start[sect];
				eaddrw = BANK_NEXT_WORD_ADDR(addrw);

				sectdone = 1;
				secterr = 0;

				while (addrw < eaddrw) {
					flash_word_t stat = *addrw;

					if ((stat & BANK_STAT_RDY)
					    != BANK_STAT_RDY)
						sectdone = 0;
					else if ((stat & BANK_STAT_ERR) != 0) {
						printf(" failed on sector %d "
							"(stat = 0x%08lx) at "
							"address 0x%08lx\n",
							sect, stat,
							(unsigned long)addrw);
						*addrw = BANK_CMD_CLR_STAT;
						secterr = 1;
					}

					addrw++;
				}

				if (secterr)
					errcount++;

				if (sectdone || secterr)
					done[sect] = 1;
				else
					alldone = 0;
			}
		}

	} while (!alldone);

	if (errcount > 0)
		printf (" failed (%d sector%s had errors)\n",
			errcount, errcount > 1 ? "s" : "");
	else
		printf (" done\n");

#else /* CFG_CMA302_PARALLEL_ERASE */

	/* erase each unprotected sector sequentially */

	start = get_timer (0);
	last = 0;
	errcount = 0;

	for (sect = s_first; sect <= s_last; sect++) {
		if (info->protect[sect] == 0) {	/* not protected */
			flash_addr_t addrw, saddrw, eaddrw;
			int sectdone;
			ulong estart;

			saddrw = (flash_addr_t)info->start[sect];
			eaddrw = BANK_NEXT_WORD_ADDR(saddrw);

#ifdef FLASH_DEBUG
			printf("erasing sector %d, start addr = 0x%08lx "
				"(bank next word addr = 0x%08lx)\n", sect,
				(unsigned long)saddrw, (unsigned long)eaddrw);
#endif

			/* Disable intrs which might cause a timeout here */
			flag = disable_interrupts();

			for (addrw = saddrw; addrw < eaddrw; addrw++) {
#ifdef FLASH_DEBUG
				printf("  writing erase cmd to addr 0x%08lx\n",
					(unsigned long)addrw);
#endif
				*addrw = BANK_CMD_ERASE1;
				*addrw = BANK_CMD_ERASE2;
			}

			/* re-enable interrupts if necessary */
			if (flag)
				enable_interrupts();

			/* wait at least 80us - let's wait 1 ms */
			udelay (1000, bd->bi_intfreq);

			estart = get_timer(start);

			do {
				now = get_timer(start);

				if (now - estart > CFG_FLASH_ERASE_TOUT) {
					printf ("Timeout (sect %d, "
						"addr 0x%08lx)\n", sect,
						(unsigned long)addrw);
					errcount = 1;
					break;
				}

#ifndef FLASH_DEBUG
				/* show that we're waiting */
				if ((now - last) > 1000) { /* every second */
					serial_putc ('.');
					last = now;
				}
#endif

				sectdone = 1;

				for (addrw = saddrw; addrw < eaddrw; addrw++) {
					flash_word_t stat = *addrw;

#ifdef FLASH_DEBUG
					printf("  checking status at addr "
						"0x%08lx [0x%08lx]\n",
						(unsigned long)addrw, stat);
#endif
					if ((stat & BANK_STAT_RDY)
					    != BANK_STAT_RDY)
						sectdone = 0;
					else if ((stat & BANK_STAT_ERR) != 0) {
						printf(" failed on sector %d "
							"(stat = 0x%08lx) at "
							"address 0x%08lx\n",
							sect, stat,
							(unsigned long)addrw);
						*addrw = BANK_CMD_CLR_STAT;
						errcount = 1;
					}
				}

			} while (!sectdone);

			if (errcount)
				break;
		}
	}

	if (errcount > 0)
		printf (" failed\n");
	else
		printf (" done\n");

#endif /* CFG_CMA302_PARALLEL_ERASE */

	/* reset to read mode */
	for (sect = s_first; sect <= s_last; sect++) {
		if (info->protect[sect] == 0) {	/* not protected */
			flash_addr_t addrw, eaddrw;

			addrw = (flash_addr_t)info->start[sect];
			eaddrw = BANK_NEXT_WORD_ADDR(addrw);

			while (addrw < eaddrw) {
#ifdef FLASH_DEBUG
				printf("  writing reset cmd to addr 0x%08lx\n",
					(unsigned long)addrw);
#endif
				*addrw = BANK_CMD_RST;
				addrw++;
			}
		}
	}
}

/*-----------------------------------------------------------------------
 */

flash_info_t *
addr2info(ulong addr)
{
	flash_info_t *info;
	int i;

	for (i=0, info=&flash_info[0]; i<CFG_MAX_FLASH_BANKS; ++i, ++info) {
		if ((addr >= info->start[0]) &&
		    (addr < (info->start[0] + info->size)) ) {
			return (info);
		}
	}

	return (NULL);
}

/*-----------------------------------------------------------------------
 * Copy memory to flash.
 * Make sure all target addresses are within Flash bounds,
 * and no protected sectors are hit.
 * Returns:
 * 0 - OK
 * 1 - write timeout
 * 2 - Flash not erased
 * 4 - target range includes protected sectors
 * 8 - target address not in Flash memory
 */
int
flash_write (uchar *src, ulong addr, ulong cnt)
{
	int i;
	ulong         end        = addr + cnt - 1;
	flash_info_t *info_first = addr2info (addr);
	flash_info_t *info_last  = addr2info (end );
	flash_info_t *info;

	if (cnt == 0) {
		return (0);
	}

	if (!info_first || !info_last) {
		return (8);
	}

	for (info = info_first; info <= info_last; ++info) {
		ulong b_end = info->start[0] + info->size;	/* bank end addr */
		short s_end = info->sector_count - 1;
		for (i=0; i<info->sector_count; ++i) {
			ulong e_addr = (i == s_end) ? b_end : info->start[i + 1];

			if ((end >= info->start[i]) && (addr < e_addr) &&
			    (info->protect[i] != 0) ) {
				return (4);
			}
		}
	}

	/* finally write data to flash */
	for (info = info_first; info <= info_last && cnt>0; ++info) {
		ulong len;
		
		len = info->start[0] + info->size - addr;
		if (len > cnt)
			len = cnt;
		if ((i = write_buff(info, src, addr, len)) != 0) {
			return (i);
		}
		cnt  -= len;
		addr += len;
		src  += len;
	}
	return (0);
}

/*-----------------------------------------------------------------------
 * Copy memory to flash, returns:
 * 0 - OK
 * 1 - write timeout
 * 2 - Flash not erased
 * 3 - write error
 */

static int
write_buff(flash_info_t *info, uchar *src, ulong addr, ulong cnt)
{
	ulong cp, wp, data;
	int i, l, rc;
	ulong start, now, last;

	wp = (addr & ~3);	/* get lower word aligned address */

	/*
	 * handle unaligned start bytes
	 */
	if ((l = addr - wp) != 0) {
		data = 0;
		for (i=0, cp=wp; i<l; ++i, ++cp) {
			data = (data << 8) | (*(uchar *)cp);
		}
		for (; i<4 && cnt>0; ++i) {
			data = (data << 8) | *src++;
			--cnt;
			++cp;
		}
		for (; cnt==0 && i<4; ++i, ++cp) {
			data = (data << 8) | (*(uchar *)cp);
		}

		if ((rc = write_word(info, wp, data)) != 0) {
			return (rc);
		}
		wp += 4;
	}

	/*
	 * handle word aligned part
	 */
	start = get_timer (0);
	last = 0;
	while (cnt >= 4) {
		data = 0;
		for (i=0; i<4; ++i) {
			data = (data << 8) | *src++;
		}
		if ((rc = write_word(info, wp, data)) != 0) {
			return (rc);
		}
		wp  += 4;
		cnt -= 4;

		/* show that we're waiting */
		now = get_timer(start);
		if ((now - last) > 1000) {	/* every second */
			serial_putc ('.');
			last = now;
		}
	}

	if (cnt == 0) {
		return (0);
	}

	/*
	 * handle unaligned tail bytes
	 */
	data = 0;
	for (i=0, cp=wp; i<4 && cnt>0; ++i, ++cp) {
		data = (data << 8) | *src++;
		--cnt;
	}
	for (; i<4; ++i, ++cp) {
		data = (data << 8) | (*(uchar *)cp);
	}
	
	return (write_word(info, wp, data));
}

/*-----------------------------------------------------------------------
 * Write a word to Flash, returns:
 * 0 - OK
 * 1 - write timeout
 * 2 - Flash not erased
 * 3 - write error
 */
static int
write_word(flash_info_t *info, ulong dest, ulong data)
{
	flash_addr_t addr = (flash_addr_t)dest;
	flash_word_t value = (flash_word_t)data, stat;
	ulong start;
	int flag, retval;

	/* Check if Flash is (sufficiently) erased */
	if ((*addr & value) != value) {
		return (2);
	}
	/* Disable interrupts which might cause a timeout here */
	flag = disable_interrupts();

	*addr = BANK_CMD_PROG;

	*addr = data;

	/* re-enable interrupts if necessary */
	if (flag)
		enable_interrupts();

	retval = 0;

	/* data polling for D7 */
	start = get_timer (0);
	do {
		if (get_timer(start) > CFG_FLASH_WRITE_TOUT) {
			retval = 1;
			goto done;
		}
		stat = *addr;
	} while ((stat & BANK_STAT_RDY) != BANK_STAT_RDY);

	if ((stat & BANK_STAT_ERR) != 0) {
		printf("flash program failed (stat = 0x%08lx) "
			"at address 0x%08lx\n", stat, dest);
		*addr = BANK_CMD_CLR_STAT;
		retval = 3;
	}

done:
	/* reset to read mode */
	*addr = BANK_CMD_RST;

	return (retval);
}

/*-----------------------------------------------------------------------
 */
